/*
 * Copyright 2017-2018 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource;

import java.io.IOException;
import java.io.InputStream;
import java.net.Inet4Address;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.text.StrMatcher;
import org.apache.commons.lang3.text.StrTokenizer;
import org.apache.commons.lang3.tuple.Pair;
import org.diirt.datasource.DataSourceConfiguration;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.aquenos.epics.jackie.client.ChannelAccessClientConfiguration;
import com.aquenos.epics.jackie.client.beacon.BeaconDetectorConfiguration;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolverConfiguration;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.JavaUtilLoggingErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessageCodec;
import com.aquenos.epics.jackie.common.util.EnvironmentUtil;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;

/**
 * <p>
 * Configuration for the {@link JackieDataSource}. This class parses the
 * configuration file in a way so that a configuration file originally written
 * for the JCA-based data source is accepted.
 * </p>
 * 
 * <p>
 * The EPICS Jackie client usually gets its configuration from the standard
 * <code>EPICS_CA_*</code> environment variables. Configuring the library
 * through environment variables is the preferred method. However, for the few
 * configuration options that cannot be set through environment variables or if
 * using environment variables is not feasible (e.g. because different
 * configuration shall be used within the same JVM), a configuration can be read
 * from an XML file or it can be initialized programmatically.
 * </p>
 * 
 * <p>
 * When creating a data source with the {@link AbstractJackieDataSourceProvider}
 * and there is a file <code>datasources/&lt;name&gt;/&lt;name&gt;.xml</code>
 * (where <code>&lt;name&gt;</code> is the name of the provider, e.g.
 * <code>ca</code> or <code>jackie</code>) in the DIIRT configuration directory,
 * the configuration is initialized using that file and only settings not
 * specified in that file are read from the environment.
 * </p>
 * 
 * <p>
 * When creating a configuration programmatically, setting a value to
 * <code>null</code> has the effect that its corresponding environment variable
 * is used. The default value for the setting is only used if the environment
 * variable is not set or if there is no environment variable corresponding to
 * the setting.
 * </p>
 * 
 * <table summary="List of configuration options">
 * <tr>
 * <th>Environment variable</th>
 * <th>Method name</th>
 * <th>XML path</th>
 * <th>Default value</th>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_ADDR_LIST</code></td>
 * <td>{@link #addressList(Collection)}</td>
 * <td>/ca/jcaContext/@addr_list</td>
 * <td><em>empty</em></td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_AUTO_ADDR_LIST</code></td>
 * <td>{@link #autoAddressList(Boolean)}</td>
 * <td>/ca/jcaContext/@auto_addr_list</td>
 * <td><code>true</code></td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_AUTO_ARRAY_BYTES</code></td>
 * <td>{@link #maxArrayBytes}</td>
 * <td>/ca/jcaContext/@auto_array_bytes</td>
 * <td><code>true</code></td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_CONN_TMO</code></td>
 * <td>{@link #connectionTimeout(Double)}</td>
 * <td>/ca/jcaContext/@connection_timeout</td>
 * <td>30.0</td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_MAX_ARRAY_BYTES</code></td>
 * <td>{@link #maxArrayBytes}</td>
 * <td>/ca/jcaContext/@max_array_bytes</td>
 * <td><em>n/a</em></td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_MAX_SEARCH_PERIOD</code></td>
 * <td>{@link #maxSearchPeriod(Double)}</td>
 * <td>/ca/jcaContext/@max_search_period</td>
 * <td>300.0</td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_MCAST_TTL</code></td>
 * <td>{@link #multicastTtl(Integer)}</td>
 * <td>/ca/jcaContext/@mcast_ttl</td>
 * <td>1</td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_NAME_SERVERS</code></td>
 * <td>{@link #nameServers(Collection)}</td>
 * <td>/ca/jcaContext/@name_servers</td>
 * <td><em>empty</em></td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_REPEATER_PORT</code></td>
 * <td>{@link #repeaterPort(Integer)}</td>
 * <td>/ca/jcaContext/@repeater_port</td>
 * <td>5065</td>
 * </tr>
 * <tr>
 * <td><code>EPICS_CA_SERVER_PORT</code></td>
 * <td>{@link #serverPort(Integer)}</td>
 * <td>/ca/jcaContext/@server_port</td>
 * <td>5064</td>
 * </tr>
 * <tr>
 * <td><em>OS dependent</em></td>
 * <td>{@link #hostname(String)}</td>
 * <td>/ca/jcaContext/@hostname</td>
 * <td><em>n/a</em></td>
 * </tr>
 * <tr>
 * <td><em>OS dependent</em></td>
 * <td>{@link #username(String)}</td>
 * <td>/ca/jcaContext/@username</td>
 * <td><em>n/a</em></td>
 * </tr>
 * <tr>
 * <td><em>n/a</em></td>
 * <td>{@link #cidBlockReuseTime(Long)}</td>
 * <td>/ca/jcaContext/@cid_block_reuse_time</td>
 * <td>900000</td>
 * </tr>
 * <tr>
 * <td><em>n/a</em></td>
 * <td>{@link #dbePropertySupported(Boolean)}</td>
 * <td>/ca/dataSourceOptions/@dbePropertySupported</td>
 * <td><code>true</code></td>
 * </tr>
 * <tr>
 * <td><em>n/a</em></td>
 * <td>{@link #honorZeroPrecision(Boolean)}</td>
 * <td>/ca/dataSourceOptions/@honorZeroPrecision</td>
 * <td><code>true</code></td>
 * </tr>
 * <tr>
 * <td><em>n/a</em></td>
 * <td>{@link #monitorMask(ChannelAccessEventMask)}</td>
 * <td>/ca/dataSourceOptions/@monitorMask</td>
 * <td><code>DBE_VALUE</code> | <code>DBE_ALARM</code></td>
 * </tr>
 * <tr>
 * <td><em>n/a</em></td>
 * <td>{@link #rtypValueOnly(Boolean)}</td>
 * <td>/ca/dataSourceOptions/@rtypValueOnly</td>
 * <td><code>false</code></td>
 * </tr>
 * </table>
 * 
 * <p>
 * The hostname and username differ from the other options because the way how
 * their value is determined if they are not set explicitly, depends on the
 * platform. In general, the library checks various different sources (including
 * Java system properties, environment variables, and the DNS) in order to
 * retrieve the real hostname and username, so that the behavior should be as
 * close to the C library as reasonably possible.
 * </p>
 * 
 * <p>
 * Instances of this class are immutable and thus inherently thread-safe.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class JackieDataSourceConfiguration
        extends DataSourceConfiguration<JackieDataSource> implements Cloneable {

    private static final ChannelAccessEventMask DBE_VALUE_OR_ALARM = ChannelAccessEventMask.DBE_VALUE
            .or(ChannelAccessEventMask.DBE_ALARM);

    private static Element getOptionalSingleElementByTagName(Element parent,
            String name) {
        NodeList nodeList = parent.getElementsByTagName(name);
        if (nodeList.getLength() > 1) {
            throw new IllegalArgumentException("Expected a single " + name
                    + " element below the " + parent.getTagName()
                    + " element, but found " + nodeList.getLength() + ".");
        }
        if (nodeList.getLength() == 1) {
            return (Element) nodeList.item(0);
        } else {
            return null;
        }
    }

    private static List<Pair<Inet4Address, Integer>> parseAddressAndPortList(
            String value) {
        LinkedList<Pair<Inet4Address, Integer>> addressList = new LinkedList<Pair<Inet4Address, Integer>>();
        StrTokenizer tokenizer = new StrTokenizer(value);
        tokenizer.setTrimmerMatcher(StrMatcher.splitMatcher());
        while (tokenizer.hasNext()) {
            String token = tokenizer.next();
            String addressPart;
            String portPart;
            int indexOfColon = token.indexOf(':');
            if (indexOfColon == -1) {
                addressPart = token;
                portPart = null;
            } else {
                addressPart = token.substring(0, indexOfColon);
                portPart = token.substring(indexOfColon + 1, token.length());
            }
            Inet4Address address;
            try {
                address = Inet4AddressUtil.stringToInetAddress(addressPart);
            } catch (IllegalArgumentException e) {
                // The token is not a valid IPv4 address, so we ignore it.
                continue;
            }
            int port = 0;
            if (portPart != null) {
                port = toPortNumber(portPart);
                if (port == -1) {
                    // The port number is invalid, so ignore this token.
                    continue;
                }
            }
            addressList.add(Pair.of(address, port));
        }
        return addressList;
    }

    private static int toPortNumber(String portString) {
        if (!StringUtils.isNumeric(portString) || portString.length() < 1
                || portString.length() > 5) {
            return 0;
        }
        int port = Integer.parseInt(portString, 10);
        if (port > 65535) {
            return 0;
        }
        return port;
    }

    private List<Pair<Inet4Address, Integer>> addressList;
    private Boolean autoAddressList;
    private Boolean autoArrayBytes;
    private Long cidBlockReuseTime;
    private boolean dbePropertySupported = true;
    private Double echoInterval;
    private boolean honorZeroPrecision = true;
    private String hostName;
    private final Logger logger = Logger.getLogger(getClass().getName());
    private Integer maxArrayBytes;
    private Double maxSearchPeriod;
    private Integer multicastTtl;
    private List<Pair<Inet4Address, Integer>> nameServers;
    private ChannelAccessEventMask monitorEventMask = DBE_VALUE_OR_ALARM;
    private Integer repeaterPort;
    private boolean rtypValueOnly = false;
    private Integer serverPort;
    private String userName;

    /**
     * Creates a configuration using default values.
     */
    public JackieDataSourceConfiguration() {
    }

    /**
     * Returns a new configuration with its address list set to the specified
     * value. This configuration object is not changed.
     * 
     * @param addressList
     *            new address list value (may be <code>null</code> in order to
     *            reset the value back to its default).
     * @return new configuration object.
     * @throws IllegalArgumentException
     *             if the address list contains <code>null</code> elements.
     */
    public JackieDataSourceConfiguration addressList(
            Collection<Pair<Inet4Address, Integer>> addressList) {
        if (addressList != null
                && addressList.stream().anyMatch((pair) -> pair == null
                        || pair.getLeft() == null || pair.getRight() == null)) {
            throw new IllegalArgumentException(
                    "The address list must not contain null elements.");
        }
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.addressList = (addressList == null) ? null
                : new ArrayList<Pair<Inet4Address, Integer>>(addressList);
        return newConfig;
    }

    /**
     * Returns a new configuration with its auto address list flag set to the
     * specified value. This configuration object is not changed.
     * 
     * @param autoAddressList
     *            new flag value (may be <code>null</code> in order to reset the
     *            value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration autoAddressList(
            Boolean autoAddressList) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.autoAddressList = autoAddressList;
        return newConfig;
    }

    /**
     * <p>
     * Returns a new configuration with its auto array bytes flag set to the
     * specified value. This configuration object is not changed.
     * </p>
     * 
     * <p>
     * Setting this option to <code>true</code> has no effect is the max. array
     * size is specified explicitly by calling {@link #maxArrayBytes(Integer)}
     * (or setting the corresponding configuration option a configuration file
     * that is loaded). The {@code autoArrayBytes} option is only effective when
     * {@code maxArrayBytes} has not been specified explicitly and would
     * normally be loaded from the corresponding environment variable.
     * </p>
     * 
     * @param autoArrayBytes
     *            new flag value (may be <code>null</code> in order to reset the
     *            value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration autoArrayBytes(
            Boolean autoArrayBytes) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.autoArrayBytes = autoArrayBytes;
        return newConfig;
    }

    /**
     * Returns a new configuration with its CID block-reuse time set to the
     * specified value. This configuration object is not changed.
     * 
     * @param cidBlockReuseTime
     *            new CID block-reuse time value (may be <code>null</code> in
     *            order to reset the value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration cidBlockReuseTime(
            Long cidBlockReuseTime) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.cidBlockReuseTime = cidBlockReuseTime;
        return newConfig;
    }

    /**
     * Returns a new configuration with its connection timeout set to the
     * specified value. This configuration object is not changed.
     * 
     * @param connectionTimeout
     *            new connection timeout value (may be <code>null</code> in
     *            order to reset the value back to its default).
     * @return new configuration object.
     * @throws IllegalArgumentException
     *             if the connection timeout is NaN.
     */
    public JackieDataSourceConfiguration connectionTimeout(
            Double connectionTimeout) {
        if (Double.isNaN(connectionTimeout)) {
            throw new IllegalArgumentException(
                    "Connection timeout must not be NaN.");
        }
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.echoInterval = Math.max(0.1, connectionTimeout);
        return newConfig;
    }

    @Override
    public JackieDataSource create() {
        return new JackieDataSource(this);
    }

    /**
     * Returns a new configuration with its DBE_PROPERTY supported flag set to
     * the specified value. This configuration object is not changed.
     * 
     * @param dbePropertySupported
     *            new flag value (may be <code>null</code> in order to reset the
     *            value back to its default).
     * @return new configuration object.
     * @see #isDbePropertySupported()
     */
    public JackieDataSourceConfiguration dbePropertySupported(
            Boolean dbePropertySupported) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.dbePropertySupported = (dbePropertySupported == null) ? true
                : dbePropertySupported;
        return newConfig;
    }

    /**
     * Returns a new configuration with its honor-zero-precision flag set to the
     * specified value. This configuration object is not changed.
     * 
     * @param honorZeroPrecision
     *            new flag value (may be <code>null</code> in order to reset the
     *            value back to its default).
     * @return new configuration object.
     * @see #isHonorZeroPrecision()
     */
    public JackieDataSourceConfiguration honorZeroPrecision(
            Boolean honorZeroPrecision) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.honorZeroPrecision = (honorZeroPrecision == null) ? true
                : honorZeroPrecision;
        return newConfig;
    }

    /**
     * Returns a new configuration with its hostname set to the specified value.
     * This configuration object is not changed.
     * 
     * @param hostname
     *            new hostname (may be <code>null</code> in order to reset the
     *            value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration hostname(String hostname) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.hostName = hostname;
        return newConfig;
    }

    /**
     * Tells whether to create a meta-data monitor using the
     * {@code DBE_PROPERTY} event mask. If <code>true</code>, each channel gets
     * a second monitor using the {@code DBE_PROPERTY} event mask, so that the
     * meta-data is updated when it changes. If <code>false</code>, the
     * meta-data is only read once when the channel is connected. The default is
     * <code>true</code>.
     * 
     * @return <code>true</code> if meta-data should be monitored using the
     *         {@code DBE_PROPERTY} event mask, <code>false</code> otherwise.
     */
    public boolean isDbePropertySupported() {
        return dbePropertySupported;
    }

    /**
     * Tells whether a precision of zero should actually be used when formatting
     * floating point numbers. If <code>true</code> a value with a precision of
     * zeros is formatted without any decimal digits. If <code>false</code> it
     * is formatted using the default number of digits for floating point
     * numbers.
     * 
     * @return <code>true</code> if a precision of zero shall be considered
     *         valid, <code>false</code> if it shall be considered to mean "use
     *         default precision".
     */
    public boolean isHonorZeroPrecision() {
        return honorZeroPrecision;
    }

    /**
     * Tells whether channels with names ending with ".RTYP" should receive
     * special treatment. If <code>true</code> the monitor created for such a
     * channel uses the type DBR_STRING instead of DBR_TIME_STRING. If
     * <code>false</code>, a regular DBR_TIME_STRING monitor is used. This
     * option exists to work around a bug in some version of EPICS where
     * monitoring the RTYP field with DBR_TIME_STRING leads to invalid data
     * being returned.
     * 
     * @return <code>true</code> if monitors for the RTYP field should receive
     *         special treatment, <code>false</code> otherwise.
     */
    public boolean isRtypValueOnly() {
        return rtypValueOnly;
    }

    /**
     * Returns a new configuration with its max. array bytes value set to the
     * specified value. Calling this method with a non-null argument always sets
     * the max. array bytes value, even if the
     * <code>EPICS_CA_MAX_ARRAY_BYTES</code> environment variable would be
     * ignored because <code>EPICS_CA_AUTO_ARRAY_BYTES</code> is not set to
     * <code>NO</code> (and {@link #autoArrayBytes(Boolean)} has not been called
     * with a value of <code>false</code>).
     * 
     * @param maxArrayBytes
     *            new max. array bytes value (may be <code>null</code> in order
     *            to reset the value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration maxArrayBytes(Integer maxArrayBytes) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.maxArrayBytes = Math.min(2147483616,
                Math.max(16384, maxArrayBytes));
        return newConfig;
    }

    /**
     * Returns a new configuration with its max. search period set to the
     * specified value. This configuration object is not changed.
     * 
     * @param maxSearchPeriod
     *            new max. search period value (may be <code>null</code> in
     *            order to reset the value back to its default).
     * @return new configuration object.
     * @throws IllegalArgumentException
     *             if max. search period is NaN.
     */
    public JackieDataSourceConfiguration maxSearchPeriod(
            Double maxSearchPeriod) {
        if (Double.isNaN(maxSearchPeriod)) {
            throw new IllegalArgumentException(
                    "Max. search period must not be NaN.");
        }
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.maxSearchPeriod = Math.max(0.1, maxSearchPeriod);
        return newConfig;
    }

    /**
     * Returns a new configuration with its multicast TTL set to the specified
     * value. This configuration object is not changed.
     * 
     * @param multicastTtl
     *            new multicast TTL value (may be <code>null</code> in order to
     *            reset the value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration multicastTtl(Integer multicastTtl) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.multicastTtl = Math.max(1, Math.min(255, multicastTtl));
        return newConfig;
    }

    /**
     * Returns a new configuration with its name servers list set to the
     * specified value. This configuration object is not changed.
     * 
     * @param nameServers
     *            new address list value (may be <code>null</code> in order to
     *            reset the value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration nameServers(
            Collection<Pair<Inet4Address, Integer>> nameServers) {
        if (nameServers != null
                && nameServers.stream().anyMatch((pair) -> pair == null
                        || pair.getLeft() == null || pair.getRight() == null)) {
            throw new IllegalArgumentException(
                    "The address list must not contain null elements.");
        }
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.nameServers = (nameServers == null) ? null
                : new ArrayList<Pair<Inet4Address, Integer>>(nameServers);
        return newConfig;
    }

    /**
     * Returns a new configuration with its monitor mask set to the specified
     * value. This configuration object is not changed.
     * 
     * @param monitorMask
     *            new monitor mask value (may be <code>null</code> in order to
     *            reset the value back to its default).
     * @return new configuration object.
     * @throws IllegalArgumentException
     *             if the mask is {@code DBE_NONE}.
     */
    public JackieDataSourceConfiguration monitorMask(
            ChannelAccessEventMask monitorMask) {
        if (monitorMask == ChannelAccessEventMask.DBE_NONE) {
            throw new IllegalArgumentException(
                    "Monitor mask must not be DBE_NONE.");
        }
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.monitorEventMask = monitorMask;
        return newConfig;
    }

    /**
     * Returns a new configuration that is the result of this configuration
     * being updated with the settings from the specified document. This
     * configuration object is not changed.
     * 
     * @param document
     *            document from which the new settings are read.
     * @return new configuration object with updated settings.
     * @throws IllegalArgumentException
     *             if the specified document is not a valid configuration.
     */
    public JackieDataSourceConfiguration read(Document document) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.readInternal(document);
        return newConfig;
    }

    @Override
    public DataSourceConfiguration<JackieDataSource> read(InputStream stream) {
        try {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory
                    .newInstance();
            DocumentBuilder documentBuilder = documentBuilderFactory
                    .newDocumentBuilder();
            Document document = documentBuilder.parse(stream);
            return read(document);
        } catch (IOException | ParserConfigurationException | SAXException e) {
            RuntimeException rte = new RuntimeException(
                    "Could not load the configuration for the JackieDataSource"
                            + (e.getMessage() != null ? (": " + e.getMessage())
                                    : "."));
            logger.throwing(JackieDataSourceConfiguration.class.getName(),
                    "read(InputStream)", rte);
            throw rte;
        }
    }

    /**
     * Returns a new configuration with its repeater port set to the specified
     * value. This configuration object is not changed.
     * 
     * @param repeaterPort
     *            new repeater port value (may be <code>null</code> in order to
     *            reset the value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration repeaterPort(Integer repeaterPort) {
        if (repeaterPort != null && repeaterPort < 1 || repeaterPort > 65535) {
            repeaterPort = null;
        }
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.repeaterPort = repeaterPort;
        return newConfig;
    }

    /**
     * Returns a new configuration with its RTYP value-only flag set to the
     * specified value. This configuration object is not changed.
     * 
     * @param rtypValueOnly
     *            new flag value (may be <code>null</code> in order to reset the
     *            value back to its default).
     * @return new configuration object.
     * @see #isRtypValueOnly()
     */
    public JackieDataSourceConfiguration rtypValueOnly(Boolean rtypValueOnly) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.rtypValueOnly = (rtypValueOnly == null) ? false
                : rtypValueOnly;
        return newConfig;
    }

    /**
     * Returns a new configuration with its server port set to the specified
     * value. This configuration object is not changed.
     * 
     * @param serverPort
     *            new server port value (may be <code>null</code> in order to
     *            reset the value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration serverPort(Integer serverPort) {
        if (serverPort != null && serverPort < 1 || serverPort > 65535) {
            serverPort = null;
        }
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.serverPort = serverPort;
        return newConfig;
    }

    /**
     * Return the client configuration generated from this configuration. The
     * client configuration is initialized with the values that have been set in
     * this configuration.
     * 
     * @return configuration for the Channel Access client.
     */
    public ChannelAccessClientConfiguration toChannelAccessClientConfiguration() {
        ErrorHandler errorHandler = new JavaUtilLoggingErrorHandler();
        BeaconDetectorConfiguration beaconDetectorConfiguration = new BeaconDetectorConfiguration(
                errorHandler, serverPort, repeaterPort);
        // The address lists that we have stored may contain invalid port
        // numbers. We want to replace all these port numbers with the
        // configured server port.
        LinkedHashSet<Pair<Inet4Address, Integer>> tcpNameServers;
        if (this.nameServers != null) {
            tcpNameServers = this.nameServers.stream()
                    .map((pair) -> Pair.of(pair.getLeft(),
                            (pair.getRight() < 1 || pair.getRight() > 65355)
                                    ? serverPort
                                    : pair.getRight()))
                    .collect(LinkedHashSet<Pair<Inet4Address, Integer>>::new,
                            LinkedHashSet<Pair<Inet4Address, Integer>>::add,
                            LinkedHashSet<Pair<Inet4Address, Integer>>::addAll);
        } else {
            tcpNameServers = null;
        }
        LinkedHashSet<Pair<Inet4Address, Integer>> udpNameServers;
        if (this.addressList != null) {
            udpNameServers = this.addressList.stream()
                    .map((pair) -> Pair.of(pair.getLeft(),
                            (pair.getRight() < 1 || pair.getRight() > 65355)
                                    ? serverPort
                                    : pair.getRight()))
                    .collect(LinkedHashSet<Pair<Inet4Address, Integer>>::new,
                            LinkedHashSet<Pair<Inet4Address, Integer>>::add,
                            LinkedHashSet<Pair<Inet4Address, Integer>>::addAll);
        } else {
            udpNameServers = null;
        }
        Integer maxArrayBytes = this.maxArrayBytes;
        if (maxArrayBytes == null && autoArrayBytes != null) {
            if (autoArrayBytes.booleanValue()) {
                // If autoArrayBytes is set to true, we have to specify
                // maxArrayBytes as the max. allowed value. Otherwise, a value
                // of NO for the EPICS_CA_AUTO_ARRAY_BYTES environment variable
                // may result in the EPICS_CA_MAX_ARRAY_BYTES environment
                // variable being used, which is not want we want if
                // autoArrayBytes is true.
                maxArrayBytes = ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE;
            } else {
                // If autoArrayBytes is set to false, we have to specify
                // maxArrayBytes as the value provided by the
                // EPICS_CA_MAX_ARRAY_SIZE environment variable (or the default
                // value of 16384 if that variable is not set). If we specified
                // null and the EPICS_CA_AUTO_ARRAY_SIZE variable is set to NO,
                // we would end up using the max. possible payload size,
                // effectiveley ignoring the autoArraySize option.
                maxArrayBytes = EnvironmentUtil.integerFromEnvironment(
                        "EPICS_CA_MAX_ARRAY_SIZE",
                        ChannelAccessMessageCodec.MIN_MAX_PAYLOAD_SIZE,
                        ChannelAccessMessageCodec.MIN_MAX_PAYLOAD_SIZE,
                        ChannelAccessMessageCodec.MAX_MAX_PAYLOAD_SIZE);
            }
        }
        ChannelNameResolverConfiguration resolverConfiguration = new ChannelNameResolverConfiguration(
                null, errorHandler, hostName, userName, serverPort,
                tcpNameServers, udpNameServers, autoAddressList,
                maxSearchPeriod, echoInterval, multicastTtl);
        ChannelAccessClientConfiguration clientConfiguration = new ChannelAccessClientConfiguration(
                null, hostName, userName, maxArrayBytes, maxArrayBytes,
                echoInterval, cidBlockReuseTime, monitorEventMask, errorHandler,
                beaconDetectorConfiguration, resolverConfiguration);
        return clientConfiguration;
    }

    /**
     * Returns a new configuration with its username set to the specified value.
     * This configuration object is not changed.
     * 
     * @param username
     *            new username (may be <code>null</code> in order to reset the
     *            value back to its default).
     * @return new configuration object.
     */
    public JackieDataSourceConfiguration username(String username) {
        JackieDataSourceConfiguration newConfig = clone();
        newConfig.userName = username;
        return newConfig;
    }

    @Override
    protected JackieDataSourceConfiguration clone() {
        try {
            return (JackieDataSourceConfiguration) super.clone();
        } catch (CloneNotSupportedException e) {
            // We implement the Cloneable interface, so we do not have to expect
            // an exception.
            throw new RuntimeException("Unexpected exception", e);
        }
    }

    private void readInternal(Document document) {
        Element caElement = document.getDocumentElement();
        if (caElement.getTagName() != "ca") {
            throw new IllegalArgumentException(
                    "Expected root element ca, but got "
                            + caElement.getTagName());
        }
        String configVersion = caElement.getAttribute("version");
        if (configVersion.isEmpty()) {
            throw new IllegalArgumentException(
                    "Mandatory attribute version missing on ca element.");
        }
        if (!configVersion.equals("1")) {
            throw new IllegalArgumentException(
                    "Expected configuration version 1, but got "
                            + StringEscapeUtils.escapeJava(configVersion)
                            + ".");
        }
        Element dataSourceOptionsElement = getOptionalSingleElementByTagName(
                caElement, "dataSourceOptions");
        if (dataSourceOptionsElement != null) {
            String monitorEventMaskString = dataSourceOptionsElement
                    .getAttribute("monitorMask");
            if (!monitorEventMaskString.isEmpty()) {
                switch (monitorEventMaskString.toUpperCase()) {
                case "VALUE":
                    this.monitorEventMask = DBE_VALUE_OR_ALARM;
                    break;
                case "ARCHIVE":
                    this.monitorEventMask = ChannelAccessEventMask.DBE_ARCHIVE;
                    break;
                case "ALARM":
                    this.monitorEventMask = ChannelAccessEventMask.DBE_ALARM;
                    break;
                default:
                    try {
                        this.monitorEventMask = ChannelAccessEventMask
                                .forNumericMask(Short
                                        .parseShort(monitorEventMaskString));
                    } catch (NumberFormatException e) {
                        throw new IllegalArgumentException(
                                "Illegal monitor event mask: "
                                        + monitorEventMaskString,
                                e);
                    }
                    break;
                }
            }
            String dbePropertySupportedString = dataSourceOptionsElement
                    .getAttribute("dbePropertySupported");
            // We have the isEmpty() check here because we want to treat an
            // empty value as the default and not necessarily as false.
            if (!dbePropertySupportedString.isEmpty()) {
                if (dbePropertySupportedString.equalsIgnoreCase("true")) {
                    this.dbePropertySupported = true;
                } else {
                    this.dbePropertySupported = false;
                }
            }
            String honorZeroPrecisionString = dataSourceOptionsElement
                    .getAttribute("honorZeroPrecision");
            // We have the isEmpty() check here because we want to treat an
            // empty value as the default and not necessarily as false.
            if (!honorZeroPrecisionString.isEmpty()) {
                if (honorZeroPrecisionString.equalsIgnoreCase("true")) {
                    this.honorZeroPrecision = true;
                } else {
                    this.honorZeroPrecision = false;
                }
            }
            // We do not support the pureJava attribute, but we warn if it
            // would cause the JCA-based data source to run in JNI-mode.
            String pureJavaString = dataSourceOptionsElement
                    .getAttribute("pureJava");
            if (!pureJavaString.isEmpty()
                    && !pureJavaString.equalsIgnoreCase("true")) {
                logger.info("Ignoring pureJava=\""
                        + StringEscapeUtils.escapeJava(pureJavaString)
                        + "\" in EPICS Jackie configuration. EPICS Jackie always runs in pure Java mode.");
            }
            String rtypValueOnlyString = dataSourceOptionsElement
                    .getAttribute("rtypValueOnly");
            if (!rtypValueOnlyString.isEmpty()) {
                if (rtypValueOnlyString.equalsIgnoreCase("true")) {
                    this.rtypValueOnly = true;
                } else {
                    this.rtypValueOnly = false;
                }
            }
            // We do not support the varArray attribute, but we warn if it
            // would cause the JCA-based data source to run in JNI-mode.
            String varArrayString = dataSourceOptionsElement
                    .getAttribute("varArray");
            if (!varArrayString.isEmpty()
                    && !varArrayString.equalsIgnoreCase("true")
                    && !varArrayString.equalsIgnoreCase("auto")) {
                logger.info("Ignoring varArray=\""
                        + StringEscapeUtils.escapeJava(varArrayString)
                        + "\" in EPICS Jackie configuration. EPICS Jackie always supports variable-sized arrays.");
            }
        }
        Element jcaContextElement = getOptionalSingleElementByTagName(caElement,
                "jcaContext");
        if (jcaContextElement != null) {
            NamedNodeMap attributes = jcaContextElement.getAttributes();
            for (int i = 0; i < attributes.getLength(); ++i) {
                Attr attribute = (Attr) attributes.item(i);
                String name = attribute.getName();
                String value = attribute.getValue();
                switch (name) {
                case "addr_list":
                    this.addressList = parseAddressAndPortList(value);
                    break;
                case "auto_addr_list":
                    this.autoAddressList = Boolean.parseBoolean(value);
                    break;
                case "auto_array_bytes":
                    this.autoArrayBytes = Boolean.parseBoolean(value);
                    break;
                case "cid_block_reuse_time":
                    double cidBlockReuseTimeDouble = Double.parseDouble(value);
                    if (Double.isNaN(cidBlockReuseTimeDouble)) {
                        throw new IllegalArgumentException(
                                "Invalid cid_block_reuse_time: " + value);
                    }
                    this.cidBlockReuseTime = Math
                            .round(cidBlockReuseTimeDouble * 1000.0);
                    break;
                case "connection_timeout":
                    this.echoInterval = Math.max(0.1,
                            Double.parseDouble(value));
                    if (Double.isNaN(this.echoInterval)) {
                        throw new IllegalArgumentException(
                                "Invalid connection_timeout: " + value);
                    }
                    break;
                case "hostname":
                    this.hostName = value;
                    break;
                case "max_array_bytes":
                    this.maxArrayBytes = Math.min(2147483616,
                            Math.max(16384, Integer.parseInt(value)));
                    break;
                case "max_search_period":
                    this.maxSearchPeriod = Math.max(0.1,
                            Double.parseDouble(value));
                    if (Double.isNaN(this.maxSearchPeriod)) {
                        throw new IllegalArgumentException(
                                "Invalid max_search_period: " + value);
                    }
                    break;
                case "mcast_ttl":
                    this.multicastTtl = Math.max(1,
                            Math.min(255, Integer.parseInt(value)));
                    break;
                case "name_servers":
                    this.nameServers = parseAddressAndPortList(value);
                    break;
                case "repeater_port":
                    this.repeaterPort = Integer.parseInt(value);
                    if (repeaterPort < 1 || repeaterPort > 65535) {
                        this.repeaterPort = null;
                    }
                    break;
                case "server_port":
                    this.serverPort = Integer.parseInt(value);
                    if (serverPort < 1 || repeaterPort > 65535) {
                        this.serverPort = null;
                    }
                    break;
                case "username":
                    this.userName = value;
                    break;
                default:
                    logger.info("Ignoring unsupported attribute " + name
                            + " on element jcaContext.");
                    break;
                }
            }
        }
    }

}
