/*
 * Copyright 2017-2019 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.internal;

import java.net.InetSocketAddress;
import java.nio.DoubleBuffer;
import java.nio.IntBuffer;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringEscapeUtils;
import org.diirt.datasource.ChannelHandlerWriteSubscription;
import org.diirt.datasource.ChannelWriteCallback;
import org.diirt.datasource.DataSourceTypeAdapter;
import org.diirt.datasource.MultiplexedChannelHandler;
import org.diirt.datasource.ValueCache;
import org.diirt.datasource.vtype.DataTypeSupport;
import org.diirt.util.array.CollectionNumbers;
import org.diirt.util.array.ListByte;
import org.diirt.util.array.ListDouble;
import org.diirt.util.array.ListFloat;
import org.diirt.util.array.ListInt;
import org.diirt.util.array.ListLong;
import org.diirt.util.array.ListNumber;
import org.diirt.util.array.ListShort;
import org.diirt.vtype.Array;
import org.diirt.vtype.Scalar;
import org.diirt.vtype.VByteArray;
import org.diirt.vtype.VString;

import com.aquenos.epics.jackie.client.ChannelAccessChannel;
import com.aquenos.epics.jackie.client.ChannelAccessClient;
import com.aquenos.epics.jackie.client.ChannelAccessConnectionListener;
import com.aquenos.epics.jackie.client.ChannelAccessMonitor;
import com.aquenos.epics.jackie.client.ChannelAccessMonitorListener;
import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessString;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeString;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;
import com.aquenos.epics.jackie.diirt.datasource.JackieDataSource;
import com.aquenos.epics.jackie.diirt.datasource.JackieDataSourceConfiguration;
import com.aquenos.epics.jackie.diirt.datasource.util.SimpleJsonGenerator;
import com.aquenos.epics.jackie.diirt.datasource.util.SimpleJsonParser;

/**
 * <p>
 * Channel handler using the EPICS Jackie library. Basically, this is a wrapper
 * around {@link ChannelAccessChannel}, but there is logic for dealing with
 * connects and disconnects, parsing configuration options that are part of the
 * channel name, registering monitors, converting to VTypes, etc.
 * </p>
 * 
 * <p>
 * This class is intended to be created by the {@link JackieDataSource} and most
 * of it functions are protected because the parent class implements the public
 * API and delegates to the protected methods for the data-source specific
 * logic.
 * </p>
 * 
 * <p>
 * The channel name used when creating this class can be the pure name of the
 * Channel Access channel or it can be the name of the Channel Access channel
 * with an additional options string appended. If such an options string is
 * present, it must represent a valid JSON object and must follow the actual
 * channel name, separated by a single space character. At the moment, two
 * different options are supported:
 * </p>
 * 
 * <p>
 * The <em>longString</em> option specifies how a channel that has a native
 * data-type of <code>DBR_CHAR</code> should be treated. If <em>longString</em>
 * is <em>true</em>, the value is treated as a null-terminated string, resulting
 * in a {@link VString} being returned where otherwise a {@link VByteArray}
 * would be returned. If <em>false</em>, a <code>DBR_CHAR</code> is always
 * treated as a {@code VByteArray}.
 * </p>
 * 
 * <p>
 * If the <em>longString</em> option is not specified, it is automatically
 * determined based on the channel name. If the channel name ends with
 * ".FIELD$", where <em>FIELD</em> is any string that starts with an upper-case
 * letter and consists only of upper-case letters and digits,
 * <em>longString</em> is set to <em>true</em>. Otherwise, it is set to
 * <em>false</em>.
 * </p>
 * 
 * <p>
 * The <em>putCallback</em> option specifies whether a write operation succeeds
 * immediately, or whether the client waits for the server to acknowledge the
 * write operation. By setting <em>putCallback</em> to <em>true</em>, the code
 * creating the channel effectively tells this implementation that it does not
 * really care whether the write was actually successful. As there is a
 * difference in the protocol between a put with and without a callback and most
 * other Channel Access clients use the no-callback variant by default, the
 * <em>putCallback</em> option is set to <em>false</em> by default.
 * </p>
 * 
 * <p>
 * Examples of valid options strings are:
 * </p>
 * 
 * <ul>
 * <li>{"longString":"false","putCallback:true"}</li>
 * <li>{"longString":"true","putCallback:false"}</li>
 * <li>{"putCallback":"true"}</li>
 * <li>{}</li>
 * </ul>
 * 
 * @author Sebastian Marsching
 */
public class JackieChannelHandler extends
        MultiplexedChannelHandler<JackieConnectionPayload, JackieMessagePayload> {

    private static final Pattern RECORD_FIELD_AS_LONG_STRING_PATTERN = Pattern
            .compile(".+\\.[A-Z][A-Z0-9]*\\$");
    private static final long OFFSET_EPICS_TO_UNIX_EPOCH_SECONDS = 631152000L;

    static {
        // We have to install the type support for the VType types because
        // otherwise values that we sent out are not going to be delivered.
        DataTypeSupport.install();
    }

    private String actualChannelName;
    private ChannelAccessClient client;
    private ChannelAccessChannel channel;
    private final JackieDataSourceConfiguration configuration;
    private final ChannelAccessConnectionListener connectionListener = (channel,
            nowConnected) -> {
        if (nowConnected) {
            channelConnected(channel);
        } else {
            channelDisconnected(channel);
        }
    };
    private ChannelAccessMonitor<? extends ChannelAccessControlsValue<?>> controlsMonitor;
    private final ChannelAccessMonitorListener<ChannelAccessControlsValue<?>> controlsMonitorListener = new ChannelAccessMonitorListener<ChannelAccessControlsValue<?>>() {
        @Override
        public void monitorError(
                ChannelAccessMonitor<? extends ChannelAccessControlsValue<?>> monitor,
                ChannelAccessStatus status, String message) {
            controlsMonitorException(monitor,
                    new ChannelAccessException(status, message));
        }

        @Override
        public void monitorEvent(
                ChannelAccessMonitor<? extends ChannelAccessControlsValue<?>> monitor,
                ChannelAccessControlsValue<?> value) {
            controlsMonitorEvent(monitor, value);
        }
    };
    private final Logger logger;
    private boolean longString = false;
    private boolean putCallback = false;
    // This cannot be a ChannelAccessMonitor<? extends
    // ChannelAccessTimeValue<?>> because it might actually be a
    // ChannelAccessMonitor<ChannelAccessValueOnlyString> when special handling
    // of the RTYP field is enabled.
    private ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> timeMonitor;
    private ChannelAccessMonitorListener<ChannelAccessGettableValue<?>> timeMonitorListener = new ChannelAccessMonitorListener<ChannelAccessGettableValue<?>>() {
        @Override
        public void monitorError(
                ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor,
                ChannelAccessStatus status, String message) {
            timeMonitorException(monitor,
                    new ChannelAccessException(status, message));
        }

        @Override
        public void monitorEvent(
                ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor,
                ChannelAccessGettableValue<?> value) {
            if (value.getType().isTimeType()) {
                timeMonitorEvent(monitor, (ChannelAccessTimeValue<?>) value);
            } else if (value.getType() == ChannelAccessValueType.DBR_STRING) {
                // We might receive a DBR_STRING if this channel uses the
                // special RTYP handling. In this case, we use the local time
                // and assume that there is no alarm. As an alternative, we
                // could create a value without an alarm status and time stamp,
                // but some application code might expect that there is always
                // this meta-data, so we rather generate it here.
                ChannelAccessString stringValue = (ChannelAccessString) value;
                long now = System.currentTimeMillis();
                ChannelAccessTimeString timeString = ChannelAccessValueFactory
                        .createTimeString(stringValue.getValue(),
                                client.getConfiguration().getCharset(),
                                ChannelAccessAlarmSeverity.NO_ALARM,
                                ChannelAccessAlarmStatus.NO_ALARM,
                                (int) (now / 1000L
                                        - OFFSET_EPICS_TO_UNIX_EPOCH_SECONDS),
                                (int) (now % 1000L * 1000000L));
                timeMonitorEvent(monitor, timeString);
            } else {
                timeMonitorException(monitor, new RuntimeException(
                        "Received a monitor event with an value of the unexpected type "
                                + value.getType().name() + "."));
            }
        }
    };
    // The parent class uses a ConcurrentHashMap for this purpose. Using a
    // concurrent data-structure does not make sense because it is only accessed
    // from synchronized methods. Using a map instead of a set does not make
    // sense for us either: The super implementation of addWriter does not check
    // whether the same subscription has been added before and always sends the
    // initial connection state. So we can do the same and always send the
    // exception if the channel is not writable, even if the same subscription
    // has been added before and thus has already received the exception.
    private HashSet<ChannelHandlerWriteSubscription> writeSubscriptionsAddedWhileDisconnected = new HashSet<ChannelHandlerWriteSubscription>();

    /**
     * <p>
     * Creates a channel handler for the specified channel name. The specified
     * channel name may include an JSON object at the end of the string. If this
     * JSON object is present, it is interpreted as a map of additional
     * configuration options that apply for this channel. In order to be
     * recognized, the JSON object must be right at the end of the channel name
     * and not be followed by any other characters (not even whitespace).
     * </p>
     * 
     * <p>
     * Please refer to the {@linkplain JackieChannelHandler class description}
     * for a list of supported configuration options.
     * </p>
     * 
     * @param client
     *            Jackie client that is used to create the channel.
     * @param channelName
     *            name of the Channel Access channel. Optionally, a string
     *            representing a JSON object may be appened. If such a string is
     *            present, it is interpreted as additional configuration options
     *            that apply to this channel.
     * @param configuration
     *            configuration of the data source creating this channel.
     * @param logger
     *            logger of the data-source creating this channel. As this is an
     *            internal class, error messages are logged with the logger of
     *            the data-source instead of a logger for this class.
     * @throws IllegalArgumentException
     *             if the supplied channel name contains an invalid JSON
     *             configuration string.
     */
    public JackieChannelHandler(ChannelAccessClient client, String channelName,
            JackieDataSourceConfiguration configuration, Logger logger) {
        super(channelName);
        if (client == null || configuration == null || logger == null) {
            throw new NullPointerException();
        }
        this.client = client;
        this.configuration = configuration;
        // We use the logger of the data-source. This has the advantage that log
        // messages will turn up in the expected category.
        this.logger = logger;
        parseChannelName(channelName);
    }

    @Override
    public Map<String, Object> getProperties() {
        TreeMap<String, Object> properties = new TreeMap<String, Object>(
                super.getProperties());
        JackieConnectionPayload connectionPayload = getConnectionPayload();
        properties.put("CA channel name", actualChannelName);
        if (connectionPayload != null && connectionPayload.isConnected()) {
            if (connectionPayload.isReadable()
                    && connectionPayload.isWritable()) {
                properties.put("CA access rights", "read, write");
            } else if (connectionPayload.isReadable()) {
                properties.put("CA access rights", "read");
            } else if (connectionPayload.isWritable()) {
                properties.put("CA access rights", "write");
            } else {
                properties.put("CA access rights", "none");
            }
            properties.put("CA connection state", "connected");
            properties.put("CA element count",
                    connectionPayload.getNativeCount());
            InetSocketAddress serverAddress = connectionPayload
                    .getServerAddress();
            if (serverAddress != null) {
                properties.put("CA host", serverAddress.getHostString() + ":"
                        + serverAddress.getPort());
            }
            properties.put("CA native data-type",
                    connectionPayload.getNativeDataType());
        } else {
            properties.put("CA connection state", "disconnected");
        }
        properties.put("Internal connection payload", connectionPayload);
        properties.put("Internal last message payload",
                getLastMessagePayload());
        properties.put("Option longString", longString);
        properties.put("Option putCallback", putCallback);
        return properties;
    }

    @Override
    protected synchronized void addWriter(
            ChannelHandlerWriteSubscription subscription) {
        super.addWriter(subscription);
        // If we are connected, but the PV is not writable, we notify the
        // subscription. If we are not connected yet, this will happen later,
        // when the PV is connected.
        if (isConnected()) {
            if (!isWriteConnected()) {
                subscription.getExceptionWriteFunction()
                        .writeValue(new ChannelAccessException(
                                ChannelAccessStatus.StatusMessage.ECA_NOWTACCESS
                                        .toStatus()));
            }
        } else {
            // We have to remember the writers that have been added while the
            // channel was disconnected. If we detect that the channel is not
            // writable after we it has been connected, we notify the writers
            // then.
            writeSubscriptionsAddedWhileDisconnected.add(subscription);
        }
    }

    // This method is only called while holding the lock for this channel, so
    // there is no need to mark it as synchronized.
    @Override
    protected void connect() {
        channel = client.getChannel(actualChannelName);
        channel.addConnectionListener(connectionListener);
    }

    // This method is only called while holding the lock for this channel, so
    // there is no need to mark it as synchronized.
    @Override
    protected void disconnect() {
        // This method is only called while holding the lock for this channel
        // handler, so we do not need any further synchronization.
        try {
            channel.destroy();
        } catch (ConcurrentNotificationException e) {
            // We can ignore such an exception. Our connection-state and monitor
            // listeners are designed so that they ignore notifications that
            // arrive after the channel (or the monitor) has been destroyed, so
            // we simply do not care.
        }
        channel = null;
        // Destroying the channel also destroys all monitors associated with it.
        controlsMonitor = null;
        timeMonitor = null;
        // There is no need to call processConnectionPayload(). The disconnect()
        // method is only called if no subscription is registered with this
        // channel and thus no one would be notified. The connection payload is
        // reset to null right after this method returns, so there is no need to
        // update the connection payload either.
    }

    /**
     * Finds and returns the best type adapter for the specified native element
     * count and native data-type. Returns <code>null</code> if no suitable type
     * adapter can be found.
     * 
     * @param nativeElementCount
     *            native element count of this channel.
     * @param nativeDataType
     *            native data-type of this channel.
     * @param longString
     *            <code>true</code> if a {@code DBR_CHAR} should be treated as a
     *            string, <code><code>false</code> if it should be treated as an
     *            array of bytes.
     * @return best matching type adapter or <code>null</code> if no type
     *         adapter matches.
     */
    protected DataSourceTypeAdapter<JackieConnectionPayload, JackieMessagePayload> findTypeAdapter(
            int nativeElementCount, ChannelAccessValueType nativeDataType,
            boolean longString) {
        if (nativeElementCount == 1) {
            switch (nativeDataType) {
            case DBR_CHAR:
                if (longString) {
                    return ArrayCharAsStringTypeAdapter.INSTANCE;
                } else {
                    return ScalarCharTypeAdapter.INSTANCE;
                }
            case DBR_DOUBLE:
                return ScalarDoubleTypeAdapter.INSTANCE;
            case DBR_ENUM:
                return ScalarEnumTypeAdapter.INSTANCE;
            case DBR_FLOAT:
                return ScalarFloatTypeAdapter.INSTANCE;
            case DBR_LONG:
                return ScalarLongTypeAdapter.INSTANCE;
            case DBR_SHORT:
                return ScalarShortTypeAdapter.INSTANCE;
            case DBR_STRING:
                return ScalarStringTypeAdapter.INSTANCE;
            default:
                // This is really unexpected because the native type should
                // always be one of these types and if it is not for some
                // reason, this should already have lead to an exception when
                // the connection payload was created.
                return null;

            }
        } else {
            switch (nativeDataType) {
            case DBR_CHAR:
                if (longString) {
                    return ArrayCharAsStringTypeAdapter.INSTANCE;
                } else {
                    return ArrayCharTypeAdapter.INSTANCE;
                }
            case DBR_DOUBLE:
                return ArrayDoubleTypeAdapter.INSTANCE;
            case DBR_ENUM:
                return ArrayEnumTypeAdapter.INSTANCE;
            case DBR_FLOAT:
                return ArrayFloatTypeAdapter.INSTANCE;
            case DBR_LONG:
                return ArrayLongTypeAdapter.INSTANCE;
            case DBR_SHORT:
                return ArrayShortTypeAdapter.INSTANCE;
            case DBR_STRING:
                return ArrayStringTypeAdapter.INSTANCE;
            default:
                // This is really unexpected because the native type should
                // always be one of these types and if it is not for some
                // reason, this should already have lead to an exception when
                // the connection payload was created.
                return null;

            }
        }
    }

    @Override
    protected DataSourceTypeAdapter<JackieConnectionPayload, JackieMessagePayload> findTypeAdapter(
            ValueCache<?> cache, JackieConnectionPayload connection) {
        DataSourceTypeAdapter<JackieConnectionPayload, JackieMessagePayload> typeAdapter;
        typeAdapter = findTypeAdapter(connection.getNativeCount(),
                connection.getNativeDataType(), connection.isLongString());
        // So far we have only checked that the adapter matches the connection
        // payload. Now, we also have to check that it matches the value cache.
        if (typeAdapter == null || typeAdapter.match(cache, connection) == 0) {
            return null;
        } else {
            return typeAdapter;
        }
    }

    @Override
    protected boolean isConnected(JackieConnectionPayload payload) {
        return payload.isConnected();
    }

    @Override
    protected boolean isWriteConnected(JackieConnectionPayload payload) {
        return payload.isWritable();
    }

    @Override
    protected synchronized void removeWrite(
            ChannelHandlerWriteSubscription subscription) {
        super.removeWrite(subscription);
        writeSubscriptionsAddedWhileDisconnected.remove(subscription);
    }

    @Override
    protected void write(Object newValue, ChannelWriteCallback callback) {
        try {
            // There is one step where we need to hold lock for this channel
            // handler. However, when calling the callback (which might happen
            // in-thread when the write operation has already finished when we
            // add the completion listener. For this reason, this method is not
            // synchronized and the synchronized part is in the separate method
            // guardedWrite.
            if (putCallback) {
                ListenableFuture<Void> future = guardedWriteCallback(newValue);
                future.addCompletionListener((v) -> {
                    try {
                        future.get();
                    } catch (ExecutionException e) {
                        Throwable cause = e.getCause();
                        if (cause != null) {
                            if (cause instanceof Exception) {
                                callback.channelWritten((Exception) cause);
                            } else {
                                callback.channelWritten(new RuntimeException(
                                        cause.getMessage(), cause));
                            }
                        } else {
                            callback.channelWritten(e);
                        }
                        return;
                    } catch (Exception e) {
                        callback.channelWritten(e);
                        return;
                    }
                    // We send null to indicate success.
                    callback.channelWritten(null);
                });
            } else {
                try {
                    guardedWriteNoCallback(newValue);
                } catch (Throwable e) {
                    if (e instanceof Exception) {
                        callback.channelWritten((Exception) e);
                    } else {
                        callback.channelWritten(new RuntimeException(e));
                    }
                    return;
                }
                // If the operation did not fail, we report success. Obviously,
                // this does not mean that the write actually succeeded, but by
                // setting putCallback to false, the caller has told us that it
                // does not care about whether the write was successful.
                callback.channelWritten(null);
            }
        } catch (Exception e) {
            callback.channelWritten(e);
        }
    }

    // We mark this method as synchronized because the monitor fields are also
    // accessed from the disconnect() method which may be called in a different
    // thread than this method. disconnect() is effectively synchronized because
    // it is only called from methods marked as synchronized.
    private synchronized void channelConnected(
            ChannelAccessChannel channelFromListener) {
        // If the channel instance passed to the listener is not the same
        // instance that we have here, we ignore the event. This can happen if a
        // late notification arrives after destroying the channel. In this case,
        // channel is going to be null or a new channel instance while
        // channelFromListener is going to be an old channel instance.
        if (channel != channelFromListener) {
            return;
        }
        int nativeCount;
        ChannelAccessValueType nativeDataType;
        boolean readable;
        boolean writable;
        try {
            nativeCount = channel.getNativeCount();
            nativeDataType = channel.getNativeDataType();
            readable = channel.isMayRead();
            writable = channel.isMayWrite();
        } catch (IllegalStateException e) {
            // If we get an IllegalStateException this means that the channel
            // has disconnected again. In this case, we simply do nothing -
            // everything will be done when the channel connects again.
            return;
        }
        InetSocketAddress serverAddress;
        try {
            serverAddress = channel.getServerAddress();
        } catch (IllegalStateException | UnsupportedOperationException e) {
            // We use reflective access to private classes in order to get the
            // server address, so there are many reasons why this might fail. We
            // simply ignore such a problem as the only consequence is that
            // there will be less useful debugging information.
            // The other reason why the operation might fail is because the
            // channel is now disconnected. In this case, the connection payload
            // is going to change again soon, and we do not bother either.
            serverAddress = null;
        }
        // Before updating the connection payload, we reset the message payload.
        // The data type associated with the new connection might not match the
        // data-type of the last connection, so we would try to convert a value
        // with the wrong type adapter.
        resetMessage();
        JackieConnectionPayload connectionPayload = JackieConnectionPayload
                .connected(client.getConfiguration().getCharset(),
                        configuration.isHonorZeroPrecision(), longString,
                        nativeCount, nativeDataType, readable, serverAddress,
                        writable);
        // Before overwriting the connection payload, we save the old one. We
        // will need it later when we decide whether to inform the write
        // subscriptions that the channel is not writable.
        JackieConnectionPayload oldConnectionPayload = getConnectionPayload();
        // We process the connection payload now. This ensures that when one of
        // the listener that we register below is called immediately (in this
        // thread), it will already see the updated connection payload.
        try {
            processConnection(connectionPayload);
        } catch (Throwable e) {
            logger.log(Level.SEVERE,
                    "processConnection(...) threw an exception for channel "
                            + getChannelName() + ".",
                    e);
        }
        // We always create the monitors, even if the channel is not readable.
        // In this case, the monitors will trigger an error which will be
        // converted into a corresponding message payload. This message payload
        // will then cause the DataSourceTypeAdapter to throw and notify the
        // subscriptions. This sounds more complicated than necessary, but it
        // has two advantages over notifying the subscriptions here:
        // 1) If a subscription is added a later point in time, it will also
        // receive the exception.
        // 2) Only read subscriptions are going to be notified. Even though
        // unlikely, it is possible that a channel is writable, but not
        // readable. In this case, sending the exception to "write monitors"
        // would be misleading and there only is a
        // reportExceptionToAllReadersAndWriters method, no
        // reportExceptionToAllReaders method.
        timeMonitor = channel.monitor(timeTypeForNativeType(nativeDataType));
        timeMonitor.addMonitorListener(timeMonitorListener);
        ChannelAccessValueType controlsType = controlsTypeForNativeType(
                nativeDataType);
        if (controlsType != null) {
            if (configuration.isDbePropertySupported()) {
                // We do not use the value received via this monitor, so
                // requesting more than a single element would be a waste of
                // bandwidth.
                // We always request a DBR_CTRL_* type, so we can safely cast
                // the monitor.
                @SuppressWarnings("unchecked")
                ChannelAccessMonitor<? extends ChannelAccessControlsValue<?>> tempControlsMonitor = (ChannelAccessMonitor<? extends ChannelAccessControlsValue<?>>) channel
                        .monitor(controlsType, 1,
                                ChannelAccessEventMask.DBE_PROPERTY);
                controlsMonitor = tempControlsMonitor;
                controlsMonitor.addMonitorListener(controlsMonitorListener);
            } else {
                // We save the current connection payload so that later, when
                // the listener is triggered, we can check whether it still is
                // the current connection payload. If it is not, the listener
                // was called late and the event shall be discarded.
                JackieConnectionPayload localSavedConnectionPayload = connectionPayload;
                try {
                    channel.get(controlsType, 1)
                            .addCompletionListener((future) -> {
                                ChannelAccessGettableValue<?> value;
                                try {
                                    value = future.get();
                                } catch (ExecutionException e) {
                                    if (e.getCause() != null) {
                                        controlsGetException(
                                                localSavedConnectionPayload,
                                                e.getCause());
                                    } else {
                                        controlsGetException(
                                                localSavedConnectionPayload, e);
                                    }
                                    return;
                                } catch (Throwable e) {
                                    controlsGetException(
                                            localSavedConnectionPayload, e);
                                    return;
                                }
                                // We know that we requested a DBR_CTRL_* value,
                                // so we can safely cast here.
                                controlsGetSuccess(localSavedConnectionPayload,
                                        (ChannelAccessControlsValue<?>) value);
                            });
                } catch (Throwable e) {
                    controlsGetException(localSavedConnectionPayload, e);
                }
            }
        }
        // If the PV is not writable, we report an exception to all writers. As
        // the writeConnected flag is not going to be changed, writers are not
        // going to be notified that the channel is connected. Without the
        // exception, they might wait more time for the channel to finally be
        // connected, which for them will simply not happen. The exception at
        // least gives them a chance to stop waiting.
        if (!writable) {
            // If this channel was writable when it was connected the last time,
            // we notify all writers of the problem. Otherwise, we only notify
            // those writers that have been added while we were disconnected.
            // The other writers have already been notified earlier.
            Exception exception = new ChannelAccessException(
                    ChannelAccessStatus.StatusMessage.ECA_NOWTACCESS
                            .toStatus());
            if (oldConnectionPayload == null
                    || oldConnectionPayload.isWritableWhenLastConnected()) {
                try {
                    reportExceptionToAllWriters(exception);
                } catch (Throwable t) {
                    logger.log(Level.SEVERE,
                            "reportExceptionToAllWriters(...) threw for channel "
                                    + getChannelName(),
                            t);
                }
            } else {
                for (ChannelHandlerWriteSubscription writeSubscription : writeSubscriptionsAddedWhileDisconnected) {
                    try {
                        writeSubscription.getExceptionWriteFunction()
                                .writeValue(exception);
                    } catch (Throwable t) {
                        logger.log(Level.SEVERE,
                                "Write subscription exception write function for channel "
                                        + getChannelName()
                                        + " threw an exception.",
                                t);
                    }
                }
            }
        }
        // We can now clear the set of writers that have been added while
        // disconnected.
        writeSubscriptionsAddedWhileDisconnected.clear();
    }

    // We mark this method as synchronized because the monitor fields are also
    // accessed from the disconnect() method which may be called in a different
    // thread than this method. disconnect() is effectively synchronized because
    // it is only called from methods marked as synchronized.
    private synchronized void channelDisconnected(
            ChannelAccessChannel channelFromListener) {
        // If the channel instance passed to the listener is not the same
        // instance that we have here, we ignore the event. This can happen if a
        // late notification arrives after destroying the channel. In this case,
        // channel is going to be null or a new channel instance while
        // channelFromListener is going to be an old channel instance.
        if (channel != channelFromListener) {
            return;
        }
        // If we are already in the disconnected state, we do nothing.
        // This can happen if the connection was established for a short moment,
        // but failed before we could actually switch to the connected state.
        if (!isConnected()) {
            return;
        }
        if (controlsMonitor != null) {
            try {
                controlsMonitor.destroy();
            } catch (ConcurrentNotificationException e) {
                // The monitor listener handles late notifications, so this is
                // not a problem for us and we can ignore the exception.
            }
            controlsMonitor = null;
        }
        if (timeMonitor != null) {
            try {
                timeMonitor.destroy();
            } catch (ConcurrentNotificationException e) {
                // The monitor listener handles late notifications, so this is
                // not a problem for us and we can ignore the exception.

            }
            timeMonitor = null;
        }
        JackieConnectionPayload oldConnectionPayload = getConnectionPayload();
        // This method might be called when the connection-state listener is
        // initially registered and the channel is not connected yet. In this
        // case, the existing connection payload is going to be null and thus
        // there is no sense in setting a new connection payload.
        if (oldConnectionPayload == null) {
            return;
        }
        JackieConnectionPayload connectionPayload = JackieConnectionPayload
                .disconnected(oldConnectionPayload);
        try {
            processConnection(connectionPayload);
        } catch (Throwable e) {
            logger.log(Level.SEVERE,
                    "processConnection(...) threw an exception for channel "
                            + getChannelName() + ".",
                    e);
        }
    }

    // We mark this method as synchronized because otherwise the connection
    // payload might change between us checking it and the
    // controlsMonitorException method being called.
    private void controlsGetException(
            JackieConnectionPayload localSavedConnectionPayload, Throwable e) {
        // If the connection payload has changed in the meantime, we do not want
        // to process this event any longer.
        if (getConnectionPayload() != localSavedConnectionPayload) {
            return;
        }
        // This method is only called if the controlsMonitor is null, so we can
        // simply pass null to controlsMonitorEvent.
        controlsMonitorException(null, e);
    }

    // We mark this method as synchronized because otherwise the connection
    // payload might change between us checking it and the controlsMonitorEvent
    // method being called.
    private synchronized void controlsGetSuccess(
            JackieConnectionPayload localSavedConnectionPayload,
            ChannelAccessControlsValue<?> value) {
        // If the connection payload has changed in the meantime, we do not want
        // to process this event any longer.
        if (getConnectionPayload() != localSavedConnectionPayload) {
            return;
        }
        // This method is only called if the controlsMonitor is null, so we can
        // simply pass null to controlsMonitorEvent.
        controlsMonitorEvent(null, value);
    }

    // We mark this method as synchronized because we use the last message
    // payload as a basis for the new one and if we did not synchronize, there
    // would be a race condition between geting the last one and setting the new
    // one. This synchronization should not have any adverse effects, because we
    // call processMessage anyway, which is already synchronized, so we simply
    // acquire the lock a bit earlier.
    private synchronized void controlsMonitorEvent(
            ChannelAccessMonitor<? extends ChannelAccessControlsValue<?>> monitorFromListener,
            ChannelAccessControlsValue<?> value) {
        // If the monitor instance passed to the listener is not the same
        // instance that we have here, we ignore the event. This can happen if a
        // late notification arrives after destroying the monitor. In this case,
        // controlsMonitor is going to be null or a new monitor instance while
        // monitorFromListener is going to be an old monitor instance.
        if (controlsMonitor != monitorFromListener) {
            return;
        }
        JackieMessagePayload oldMessagePayload = getLastMessagePayload();
        JackieMessagePayload newMessagePayload = JackieMessagePayload
                .withControlsValue(value, oldMessagePayload);
        try {
            processMessage(newMessagePayload);
        } catch (Throwable t) {
            logger.log(Level.SEVERE,
                    "processMessage(...) threw for channel " + getChannelName(),
                    t);
        }
    }

    // We mark this method as synchronized because we use the last message
    // payload as a basis for the new one and if we did not synchronize, there
    // would be a race condition between geting the last one and setting the new
    // one. This synchronization should not have any adverse effects, because we
    // call processMessage anyway, which is already synchronized, so we simply
    // acquire the lock a bit earlier.
    private synchronized void controlsMonitorException(
            ChannelAccessMonitor<? extends ChannelAccessControlsValue<?>> monitorFromListener,
            Throwable e) {
        // If the monitor instance passed to the listener is not the same
        // instance that we have here, we ignore the event. This can happen if a
        // late notification arrives after destroying the monitor. In this case,
        // controlsMonitor is going to be null or a new monitor instance while
        // monitorFromListener is going to be an old monitor instance.
        if (controlsMonitor != monitorFromListener) {
            return;
        }
        JackieMessagePayload oldMessagePayload = getLastMessagePayload();
        JackieMessagePayload newMessagePayload = JackieMessagePayload
                .withControlsException(e, oldMessagePayload);
        try {
            processMessage(newMessagePayload);
        } catch (Throwable t) {
            logger.log(Level.SEVERE,
                    "processMessage(...) threw for channel " + getChannelName(),
                    t);
        }
    }

    private ChannelAccessValueType controlsTypeForNativeType(
            ChannelAccessValueType nativeDataType) {
        // Strings do not have additional meta-data, so registering a controls
        // monitor does not make sense.
        // If this channel is configured for long-string mode and we have a
        // DBR_CHAR, there is no sense in requesting the meta-data either
        // because we are not going to use it anyway.
        if (nativeDataType == ChannelAccessValueType.DBR_STRING) {
            return null;
        } else if (longString
                && nativeDataType == ChannelAccessValueType.DBR_CHAR) {
            return null;
        } else {
            return nativeDataType.toControlsType();
        }
    }

    private ChannelAccessPuttableValue<?> convertValueForWrite(
            final Object originalValue) {
        // Sometimes, we want to convert the value in multiple steps, but we
        // always save the original value so that we can give the right error
        // message (about a problem with the original type) in case we cannot
        // successfully convert the value.
        Object value = originalValue;
        // We remember whether we already copied an array. This way, we can
        // avoid copying it twice. We make at least one copy in order to avoid
        // some calling code doing something stupid (like modifying the passed
        // data after this method returns, but before the data has actually been
        // transferred over the network).
        boolean needCopyOfArray = true;
        // We can never handle a value of null.
        if (value == null) {
            throw new NullPointerException("Cannot write value of null.");
        }
        // First, we look for some VTypes that we can easily support. We do not
        // really expect them, but supporting them is simple, so it is worth the
        // effort.
        if (value instanceof Scalar) {
            value = ((Scalar) value).getValue();
        } else if (value instanceof Array) {
            value = ((Array) value).getData();
        }
        // Next, we look for some less-expected types and convert them to the
        // ones we usually deal with.
        if (value instanceof Byte[]) {
            logger.warning("Writing a Byte[] to channel " + getChannelName()
                    + ": Use org.diirt.util.array.ListByte instead.");
            value = ArrayUtils.toPrimitive((Byte[]) value);
            needCopyOfArray = false;
        } else if (value instanceof Short[]) {
            logger.warning("Writing a Short[] to channel " + getChannelName()
                    + ": Use org.diirt.util.array.ListShort instead.");
            value = ArrayUtils.toPrimitive((Short[]) value);
            needCopyOfArray = false;
        } else if (value instanceof Integer[]) {
            logger.warning("Writing an Integer[] to channel " + getChannelName()
                    + ": Use org.diirt.util.array.ListInt instead.");
            value = ArrayUtils.toPrimitive((Integer[]) value);
            needCopyOfArray = false;
        } else if (value instanceof Long[]) {
            logger.warning("Writing a Long[] to channel " + getChannelName()
                    + ": Use org.diirt.util.array.ListLong instead.");
            value = ArrayUtils.toPrimitive((Long[]) value);
            needCopyOfArray = false;
        } else if (value instanceof Float[]) {
            logger.warning("Writing a Float[] to channel " + getChannelName()
                    + ": Use org.diirt.util.array.ListFloat instead.");
            value = ArrayUtils.toPrimitive((Float[]) value);
            needCopyOfArray = false;
        } else if (value instanceof Double[]) {
            logger.warning("Writing a Double[] to channel " + getChannelName()
                    + ": Use org.diirt.util.array.ListDouble instead.");
            value = ArrayUtils.toPrimitive((Double[]) value);
            needCopyOfArray = false;
        }
        // If we got a NumberList of a type that we know, we convert it to an
        // array.
        if (value instanceof ListByte) {
            value = CollectionNumbers.byteArrayCopyOf((ListByte) value);
            needCopyOfArray = false;
        } else if (value instanceof ListShort) {
            value = CollectionNumbers.shortArrayCopyOf((ListShort) value);
            needCopyOfArray = false;
        } else if (value instanceof ListInt) {
            value = CollectionNumbers.intArrayCopyOf((ListInt) value);
            needCopyOfArray = false;
        } else if (value instanceof ListLong) {
            value = CollectionNumbers.longArrayCopyOf((ListLong) value);
            needCopyOfArray = false;
        } else if (value instanceof ListFloat) {
            value = CollectionNumbers.floatArrayCopyOf((ListFloat) value);
            needCopyOfArray = false;
        } else if (value instanceof ListDouble) {
            value = CollectionNumbers.doubleArrayCopyOf((ListDouble) value);
            needCopyOfArray = false;
        } else if (value instanceof ListNumber) {
            value = CollectionNumbers.doubleArrayCopyOf((ListNumber) value);
            needCopyOfArray = false;
            logger.warning("Instance of " + value.getClass().getName()
                    + " written to channel " + getChannelName()
                    + " has been converted to a ListDouble, possibly losing information.");
        }
        // Now we take care of all the types that we handle directly.
        if (value instanceof Byte) {
            return ChannelAccessValueFactory
                    .createChar(new byte[] { ((Byte) value) });
        } else if (value instanceof Short) {
            return ChannelAccessValueFactory
                    .createShort(new short[] { ((Short) value) });
        } else if (value instanceof Integer) {
            return ChannelAccessValueFactory
                    .createLong(new int[] { ((Integer) value) });
        } else if (value instanceof Long) {
            long longValue = (Long) value;
            if (longValue >= Integer.MIN_VALUE
                    && longValue <= Integer.MAX_VALUE) {
                return ChannelAccessValueFactory
                        .createLong(new int[] { ((int) longValue) });
            } else {
                // If after converting from long to double and back the number
                // has changed, we know that we have lost precision and we log a
                // warning message.
                if (((long) ((double) longValue)) != longValue) {
                    logger.warning("Long value " + longValue
                            + " written to channel " + getChannelName()
                            + " has been converted to double value "
                            + ((double) longValue) + ", losing precision.");
                }
                return ChannelAccessValueFactory
                        .createDouble(new double[] { longValue });
            }
        } else if (value instanceof Float) {
            return ChannelAccessValueFactory
                    .createFloat(new float[] { ((Float) value) });
        } else if (value instanceof Double) {
            return ChannelAccessValueFactory
                    .createDouble(new double[] { ((Double) value) });
        } else if (value instanceof CharSequence) {
            boolean convertStringToWaveform = false;
            JackieConnectionPayload connectionPayload = getConnectionPayload();
            if (connectionPayload == null || !connectionPayload.isConnected()) {
                // If the channel is not connected, the write operation will
                // fail anyway, so we can fail it now.
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_DISCONN
                                .toStatus());
            }
            convertStringToWaveform = longString && connectionPayload
                    .getNativeDataType() == ChannelAccessValueType.DBR_CHAR;
            if (convertStringToWaveform) {
                // We truncate the string in order to ensure that it really
                // remains null-terminated. If it is truncated on the server
                // side, the null-termination might get lost.
                byte[] stringAsBytes = NullTerminatedStringUtil
                        .stringToTruncatedNullTerminatedFixedBytes(
                                (String) value,
                                connectionPayload.getNativeCount(),
                                client.getConfiguration().getCharset());
                return ChannelAccessValueFactory.createChar(stringAsBytes);
            } else {
                return ChannelAccessValueFactory.createString(
                        Collections.singleton(value.toString()),
                        client.getConfiguration().getCharset());

            }
        } else if (value instanceof byte[]) {
            byte[] byteArray = (byte[]) value;
            if (needCopyOfArray) {
                byteArray = byteArray.clone();
            }
            return ChannelAccessValueFactory.createChar(byteArray);
        } else if (value instanceof short[]) {
            short[] shortArray = (short[]) value;
            if (needCopyOfArray) {
                shortArray = shortArray.clone();
            }
            return ChannelAccessValueFactory.createShort(shortArray);
        } else if (value instanceof int[]) {
            int[] intArray = (int[]) value;
            if (needCopyOfArray) {
                intArray = intArray.clone();
            }
            return ChannelAccessValueFactory.createLong(intArray);
        } else if (value instanceof long[]) {
            long[] longArray = (long[]) value;
            boolean fitsIntoInt = Arrays.stream(longArray).allMatch(
                    (l) -> l >= Integer.MIN_VALUE && l <= Integer.MAX_VALUE);
            if (fitsIntoInt) {
                IntBuffer intBuffer = IntBuffer.allocate(longArray.length);
                Arrays.stream(longArray).mapToInt((l) -> ((int) l))
                        .forEach((i) -> intBuffer.put(i));
                return ChannelAccessValueFactory.createLong(intBuffer.array());
            } else {
                DoubleBuffer doubleBuffer = DoubleBuffer
                        .allocate(longArray.length);
                Arrays.stream(longArray).asDoubleStream()
                        .forEach((d) -> doubleBuffer.put(d));
                // If after converting from long to double and back the number
                // has changed, we know that we have lost precision and we log a
                // warning message.
                if (Arrays.stream(longArray)
                        .anyMatch((l) -> ((long) ((double) l)) != l)) {
                    logger.warning("Long array written to channel "
                            + getChannelName()
                            + " has been converted to double array, losing precision.");
                }

                return ChannelAccessValueFactory
                        .createDouble(doubleBuffer.array());
            }
        } else if (value instanceof float[]) {
            float[] floatArray = (float[]) value;
            if (needCopyOfArray) {
                floatArray = floatArray.clone();
            }
            return ChannelAccessValueFactory.createFloat(floatArray);
        } else if (value instanceof double[]) {
            double[] doubleArray = (double[]) value;
            if (needCopyOfArray) {
                doubleArray = doubleArray.clone();
            }
            return ChannelAccessValueFactory.createDouble(doubleArray);
        } else {
            throw new IllegalArgumentException(
                    "Value of type " + originalValue.getClass().getName()
                            + " cannot be written to channel "
                            + getChannelName() + ".");
        }
    }

    // We have to mark the write method as synchronized because it accesses the
    // channel field. In addition to that, this has the advantage that the
    // connection payload cannot change unexpectedly while we process the write
    // request (otherwise it could due to sudden disconnects or reconnects of
    // the channel.
    private synchronized ListenableFuture<Void> guardedWriteCallback(
            Object newValue) {
        ChannelAccessPuttableValue<?> caValue = convertValueForWrite(newValue);
        // We can assume that the channel is not null because this method should
        // only be called after calling connect() and before calling
        // disconnect().
        return channel.put(caValue);
    }

    // We have to mark the write method as synchronized because it accesses the
    // channel field. In addition to that, this has the advantage that the
    // connection payload cannot change unexpectedly while we process the write
    // request (otherwise it could due to sudden disconnects or reconnects of
    // the channel.
    private synchronized void guardedWriteNoCallback(Object newValue) {
        ChannelAccessPuttableValue<?> caValue = convertValueForWrite(newValue);
        // We can assume that the channel is not null because this method should
        // only be called after calling connect() and before calling
        // disconnect().
        channel.putNoCallback(caValue);
    }

    private void parseChannelName(String fullChannelName) {
        // A channel name might consist of the actual channel name followed by
        // optional parameters that configure the behavior of this data source.
        // In order to be compatible with the JCA-based data-source, we use the
        // same format for these options.
        // This means that these options are enclosed in curly braces and follow
        // a JSON-style syntax. We also use the same option names.
        // Extracting the JSON-string is a bit tricky: A valid channel name
        // might contain a space and curly braces, so we cannot simply cut at
        // the first combination of space and opening curly brace. JSON, on the
        // other hand, might contain objects within the object, so cutting at
        // the last combination of a space and opening curly brace is not
        // necessarily correct either.
        // However, channel names rarely contain spaces, so cutting at the first
        // occurrence of a space and an opening curly brace is a pretty good
        // assumption. If this does not work (the resulting string is not valid
        // JSON), we simply look for other places where we can cut.
        // If the string does not end with a closing curly brace, our life is
        // much simpler and we can simply assume that there is no JSON string at
        // the end of the channel name.
        boolean forceNoLongString = false;
        if (fullChannelName.endsWith("}")) {
            int spaceIndex = fullChannelName.indexOf(" {");
            Object jsonObj = null;
            // We remember the first exception because the first place where we
            // cut the string is most likely the right place.
            IllegalArgumentException firstException = null;
            while (spaceIndex != -1) {
                try {
                    jsonObj = SimpleJsonParser.parse(fullChannelName.substring(
                            spaceIndex + 1, fullChannelName.length()));
                    firstException = null;
                    break;
                } catch (IllegalArgumentException e) {
                    // We try a larger portion of the string, but we save the
                    // exception in case the other attempts fail as well.
                    if (firstException == null) {
                        firstException = e;
                    }
                }
                spaceIndex = fullChannelName.indexOf(" {", spaceIndex + 2);
            }
            if (firstException != null) {
                logger.warning("Ignoring JSON options in channel name \""
                        + StringEscapeUtils.escapeJava(fullChannelName)
                        + "\" because they cannot be parsed: "
                        + firstException.getMessage());
            } else if (jsonObj != null) {
                // jsonObj must be a map because we know that the string
                // represents a JSON object (because of the curly braces).
                @SuppressWarnings("unchecked")
                Map<String, Object> options = (Map<String, Object>) jsonObj;
                Object longStringOption = options.get("longString");
                if (Boolean.TRUE.equals(longStringOption)) {
                    longString = true;
                } else if (Boolean.FALSE.equals(longStringOption)) {
                    longString = false;
                    forceNoLongString = true;
                } else if (options.containsKey("longString")) {
                    throw new IllegalArgumentException(
                            "\"longString\" option in channel name \""
                                    + StringEscapeUtils
                                            .escapeJava(fullChannelName)
                                    + "\" has invalid value "
                                    + SimpleJsonGenerator
                                            .generate(longStringOption)
                                    + " (true or false was expected). The option is going to be ignored.");
                }
                Object putCallbackOption = options.get("putCallback");
                if (Boolean.TRUE.equals(putCallbackOption)) {
                    putCallback = true;
                } else if (Boolean.FALSE.equals(putCallbackOption)) {
                    putCallback = false;
                } else if (options.containsKey("putCallback")) {
                    throw new IllegalArgumentException(
                            "\"putCallback\" option in channel name \""
                                    + StringEscapeUtils
                                            .escapeJava(fullChannelName)
                                    + "\" has invalid value "
                                    + SimpleJsonGenerator.generate(putCallback)
                                    + " (true or false was expected). The option is going to be ignored.");
                }
                actualChannelName = fullChannelName.substring(0, spaceIndex);
            }
        }
        // If the actualChannelName has not been set yet, there is no valid JSON
        // options part and the full channel name is the actual channel name.
        if (actualChannelName == null) {
            actualChannelName = fullChannelName;
        }
        // When reading fields from an IOC's record, one can read them as long
        // strings (arrays of chars) by appending a dollar sign to the end of
        // their names. If we find a channel name that matches this scheme, we
        // assume that the array of chars should actually be treated as a
        // string.
        // We do not automatically set the longString option if it has been
        // explicitly set to false by the user.
        if (!longString && !forceNoLongString
                && RECORD_FIELD_AS_LONG_STRING_PATTERN
                        .matcher(actualChannelName).matches()) {
            longString = true;
        }
    }

    // We mark this method as synchronized because we use the last message
    // payload as a basis for the new one and if we did not synchronize, there
    // would be a race condition between geting the last one and setting the new
    // one. This synchronization should not have any adverse effects, because we
    // call processMessage anyway, which is already synchronized, so we simply
    // acquire the lock a bit earlier.
    private synchronized void timeMonitorEvent(
            ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorFromListener,
            ChannelAccessTimeValue<?> value) {
        // If the monitor instance passed to the listener is not the same
        // instance that we have here, we ignore the event. This can happen if a
        // late notification arrives after destroying the monitor. In this case,
        // timeMonitor is going to be null or a new monitor instance while
        // monitorFromListener is going to be an old monitor instance.
        if (timeMonitor != monitorFromListener) {
            return;
        }
        JackieMessagePayload oldMessagePayload = getLastMessagePayload();
        JackieMessagePayload newMessagePayload = JackieMessagePayload
                .withTimeValue(value, oldMessagePayload);
        try {
            processMessage(newMessagePayload);
        } catch (Throwable t) {
            logger.log(Level.SEVERE,
                    "processMessage(...) threw for channel " + getChannelName(),
                    t);
        }
    }

    // We mark this method as synchronized because we use the last message
    // payload as a basis for the new one and if we did not synchronize, there
    // would be a race condition between geting the last one and setting the new
    // one. This synchronization should not have any adverse effects, because we
    // call processMessage anyway, which is already synchronized, so we simply
    // acquire the lock a bit earlier.
    private synchronized void timeMonitorException(
            ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorFromListener,
            Throwable e) {
        // If the monitor instance passed to the listener is not the same
        // instance that we have here, we ignore the event. This can happen if a
        // late notification arrives after destroying the monitor. In this case,
        // timeMonitor is going to be null or a new monitor instance while
        // monitorFromListener is going to be an old monitor instance.
        if (timeMonitor != monitorFromListener) {
            return;
        }
        JackieMessagePayload oldMessagePayload = getLastMessagePayload();
        JackieMessagePayload newMessagePayload = JackieMessagePayload
                .withTimeException(e, oldMessagePayload);
        try {
            processMessage(newMessagePayload);
        } catch (Throwable t) {
            logger.log(Level.SEVERE,
                    "processMessage(...) threw for channel " + getChannelName(),
                    t);
        }
    }

    private ChannelAccessValueType timeTypeForNativeType(
            ChannelAccessValueType nativeDataType) {
        // If the corresponding configuration flag is enabled, we want to handle
        // the RTYP field in a special way.
        if (configuration.isRtypValueOnly()
                && nativeDataType == ChannelAccessValueType.DBR_STRING
                && actualChannelName.endsWith(".RTYP")) {
            return nativeDataType;
        }
        // In theory, it is possible that the server sends a data-type that has
        // no corresponding DBR_TIME_* type. In particular, this happens if it
        // sends a DBR_PUT_ACKT, DBR_PUT_ACKS, DBR_STSACK_STRING, or
        // DBR_CLASS_NAME. Sending a DBR_PUT_ACKT or DBR_PUT_ACKS are only used
        // in write operations and DBR_STSACK_STRING and DBR_CLASS_NAME are only
        // used in read operations when specifically requested. In fact, the CA
        // server of EPICS Base will never report such a native type and as it
        // does not make much sense, it is unlikely any other implementation
        // will. For this reason, we do not include special handling for these
        // cases and simply rethrow the exception (with a slightly more accurate
        // description of the problem).
        try {
            return nativeDataType.toTimeType();
        } catch (IllegalArgumentException e) {
            throw new RuntimeException("Unexpected native type for channel \""
                    + StringEscapeUtils.escapeJava(getChannelName()) + "\": "
                    + nativeDataType.name());
        }
    }

}
