/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.util;

import static org.junit.Assert.*;

import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import org.junit.Test;

import com.aquenos.epics.jackie.diirt.datasource.util.SimpleJsonParser;

/**
 * Tests for the {@link SimpleJsonParser}.
 * 
 * @author Sebastian Marsching
 */
public class SimpleJsonParserTest {

    private static Object parse(String jsonString) {
        return SimpleJsonParser.parse(jsonString);
    }

    private static void testNumber(String number) {
        assertEquals(Double.parseDouble(number),
                ((Number) parse(number)).doubleValue(), 0.00001);
    }

    /**
     * Tests that JSON arrays are parsed correctly.
     */
    @Test
    public void arrays() {
        assertEquals(Collections.emptyList(), parse("[]"));
        assertEquals(Collections.emptyList(), parse("[\t]"));
        assertEquals(Collections.singletonList(true), parse("[true]"));
        assertEquals(Collections.singletonList("abc"), parse("[ \"abc\"]"));
        assertEquals(Collections.singletonList(null), parse("[null ]"));
        assertEquals(Arrays.asList("abc", null, "def", false),
                parse("[ \"abc\", null,\"def\"  , false ]"));
    }

    /**
     * Test that the parsing fails if there is a comma in an empty array.
     */
    @Test(expected = IllegalArgumentException.class)
    public void commaInEmptyArrayNotAllowed() {
        parse("[,]");
    }

    /**
     * Test that the parsing fails if there is a comma in an empty array.
     */
    @Test(expected = IllegalArgumentException.class)
    public void commaInEmptyObjectNotAllowed() {
        parse("{,}");
    }

    /**
     * Tests that the JSON document "false" is parsed correctly.
     */
    @Test
    public void falseValue() {
        assertEquals(Boolean.FALSE, parse("false"));
    }

    /**
     * Test that the parsing fails if there is leading comma in an array.
     */
    @Test(expected = IllegalArgumentException.class)
    public void leadingCommaInArrayNotAllowed() {
        parse("[,\"abc\"]");
    }

    /**
     * Test that the parsing fails if there is leading comma in an object.
     */
    @Test(expected = IllegalArgumentException.class)
    public void leadingCommaInObjectNotAllowed() {
        parse("{,\"a\": 5}");
    }

    /**
     * Test that the parsing fails if there is leading whitespace.
     */
    @Test(expected = IllegalArgumentException.class)
    public void leadingWhitespaceNotAllowed() {
        parse(" 5");
    }

    /**
     * Tests that the JSON document "null" is parsed correctly.
     */
    @Test
    public void nullValue() {
        assertNull(parse("null"));
    }

    /**
     * Tests that JSON numbers are parsed correctly.
     */
    @Test
    public void numberValues() {
        testNumber("5.384");
        testNumber("-7.384");
        testNumber("2.0e-3");
        testNumber("-5e22");
        testNumber("1234567890");
        testNumber("0");
        testNumber("0.00");
        testNumber("-48");
        testNumber("1e50000");
    }

    /**
     * Tests that JSON objects are parsed correctly.
     */
    @Test
    public void objects() {
        assertEquals(Collections.emptyMap(), parse("{}"));
        assertEquals(Collections.emptyMap(), parse("{ \n}"));
        assertEquals(Collections.singletonMap("boolean", true),
                parse("{\"boolean\":true}"));
        assertEquals(Collections.singletonMap("string", "abc"),
                parse("{ \"string\"  : \"abc\" }"));
        assertEquals(Collections.singletonMap("null", null),
                parse("{\"null\": null }"));
        assertEquals(
                Collections.singletonMap("nested",
                        Collections.singletonMap("test", true)),
                parse("{\"nested\":{\"test\":true}}"));
        LinkedHashMap<String, Object> testMap = new LinkedHashMap<String, Object>();
        testMap.put("k1", "abc");
        testMap.put("k2", null);
        testMap.put("k3", "def");
        testMap.put("k4", false);
        String testJson = "{ \"k1\": \"abc\", \"k2\":null,\"k3\":  \"def\"  , \"k4\" : false }";
        @SuppressWarnings("unchecked")
        Map<String, Object> resultMap = (Map<String, Object>) parse(testJson);
        // We want to be sure that the result map has the right order, so we
        // cannot simply use assertEquals().
        assertEquals(testMap.size(), resultMap.size());
        Iterator<Map.Entry<String, Object>> i1 = testMap.entrySet().iterator();
        Iterator<Map.Entry<String, Object>> i2 = resultMap.entrySet()
                .iterator();
        while (i1.hasNext()) {
            assertEquals(i1.next(), i2.next());
        }
    }

    /**
     * Test that the parsing fails if there is an object that has a key without
     * an associated value.
     */
    @Test(expected = IllegalArgumentException.class)
    public void objectWithKeyAndNoValueNotAllowed() {
        parse("{\"a\"}");
    }

    /**
     * Tests that JSON strings are parsed correctly.
     */
    @Test
    public void stringValues() {
        assertEquals("a\"b\\c\näöü", parse("\"a\\\"b\\\\c\\näöü\""));
        assertEquals("", parse("\"\""));
        assertEquals("\"", parse("\"\\\"\""));
        assertEquals(" \n@>", parse("\" \\n\\u0040\\u003e\""));
    }

    /**
     * Test that the parsing fails if there is trailing comma in an array.
     */
    @Test(expected = IllegalArgumentException.class)
    public void trailingCommaInArrayNotAllowed() {
        parse("[5,]");
    }

    /**
     * Test that the parsing fails if there is trailing comma in an object.
     */
    @Test(expected = IllegalArgumentException.class)
    public void trailingCommaInObjectNotAllowed() {
        parse("{\"a\": 5,}");
    }

    /**
     * Test that the parsing fails if there is trailing whitespace.
     */
    @Test(expected = IllegalArgumentException.class)
    public void trailingWhitespaceNotAllowed() {
        parse("48\t");
    }

    /**
     * Tests that the JSON document "true" is parsed correctly.
     */
    @Test
    public void trueValue() {
        assertEquals(Boolean.TRUE, parse("true"));
    }

}
